/* Copyright (c) 2012 the authors listed at the following URL, and/or
the authors of referenced articles or incorporated external code:
http://en.literateprograms.org/Suffix_tree_(Java)?action=history&offset=20100123220137

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Retrieved from: http://en.literateprograms.org/Suffix_tree_(Java)?oldid=16641
*/


import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class SuffixTree {
	public void sampleUsage() {
		
		CompactSuffixTree tree = new CompactSuffixTree(new SimpleSuffixTree("bananas"));
		String properties = "rankdir=LR; node[shape=box fillcolor=gray95 style=filled]\n";
		System.out.println("digraph {\n" + properties + tree.root + "}");

	}
}
abstract class AbstractSuffixTree {
	    
	String text = null;
	SuffixTreeNode root = null;
	int inputAlphabetSize = -1;
	
	AbstractSuffixTree() {}
	
	AbstractSuffixTree(String text) {
	    if (text.length() > 0 && text.charAt(text.length() - 1) == '$') {
	        this.text = text;
	    } else {
	        this.text = text + "$";
	    }
	}
}

class SimpleSuffixTree extends AbstractSuffixTree {
	   
	public SimpleSuffixTree(String text) {
		super(text);
		constructTree();
	}
	
	public SimpleSuffixTree(List<String> texts) {
		super();
	    super.root = new SuffixTreeNode();
	    int i=0;
	    for(String text:texts){
	    	if((++i)*1000./texts.size()-(i)*1000/texts.size()==0)System.out.println((i)*1000/texts.size()+"/1000");
		    if (text.length() > 0 && text.charAt(text.length() - 1) == '$') {
		    } else {
		        text = text + "$";
		    }
	    	this.addSuffix(text);
	    }
	}
	
	
	private void constructTree() {
	    super.root = new SuffixTreeNode();
	    for (int i = 0; i < super.text.length(); i++) {
	        super.root.addSuffix(super.text.substring(i), i+1);
	    }
	}
	
	private void addSuffix(String s){
	    for (int i = 0; i < s.length(); i++) {
	        super.root.addSuffix(s.substring(i), i+1);
	    }
	}
	
	
}

class CompactSuffixTree extends AbstractSuffixTree {

	
	public CompactSuffixTree(SimpleSuffixTree simpleSuffixTree) {
	    super(simpleSuffixTree.text);
	    super.root = compactNodes(simpleSuffixTree.root, 0);
	}
	
	public CompactSuffixTree(List<String> suffixes) {
	    super.root = compactNodes(new SimpleSuffixTree(suffixes).root, 0);
		
	}

	
	private SuffixTreeNode compactNodes(SuffixTreeNode node, int nodeDepth) {
	    node.nodeDepth = nodeDepth;
	    for (SuffixTreeNode child : node.children) {
	        while (child.children.size() == 1) {
	            SuffixTreeNode grandchild = child.children.iterator().next();
	            child.incomingEdge.label += grandchild.incomingEdge.label;
	            child.stringDepth += grandchild.incomingEdge.label.length();
	            child.children = grandchild.children;
	            for (SuffixTreeNode grandchild2 : child.children)
	                grandchild2.parent = node;
	        }
	        child = compactNodes(child, nodeDepth + 1);
	    }
	    return node;
	}
	
	
	
}
class SuffixTreeNode {
	
	SuffixTreeEdge incomingEdge = null;
	int nodeDepth = -1;
	int label = -1;
	Collection<SuffixTreeNode> children = null;
	SuffixTreeNode parent = null;
	int stringDepth;
	int id = 0;
	int numberOfLeaves=1;
	public static int c;
	    
	public SuffixTreeNode(SuffixTreeNode parent, char incomingLabel,
	        int depth, int label, int id) {
	    children = new ArrayList<SuffixTreeNode>();
	    incomingEdge = new SuffixTreeEdge(incomingLabel, label);
	    nodeDepth = depth;
	    this.label = label;
	    this.parent = parent;
	    stringDepth = parent.stringDepth + 1;//incomingLabel.length();
	    this.id = id;
	}

	public SuffixTreeNode() {
	    children = new ArrayList<SuffixTreeNode>();
	    nodeDepth = 0;
	    label = 0;
	}
	
	public void addSuffix(String suffix, int pathIndex) {
	    SuffixTreeNode insertAt = this;
	    insertAt = search(this, suffix);
	    insert(insertAt, suffix, pathIndex);
	}
	

	
	public SuffixTreeNode search(SuffixTreeNode startNode, String suffix) {
	    if (suffix.isEmpty()) {
	        throw new IllegalArgumentException(
	                "Empty suffix. Probably no valid simple suffix tree exists for the input.");
	    }
	    Collection<SuffixTreeNode> children = startNode.children;
	    for (SuffixTreeNode child : children) {
	        if (child.incomingEdge.label.charAt(0)==suffix.charAt(0)) {
	           suffix= suffix.substring(1);
	            if (suffix.isEmpty()) {
			        startNode.numberOfLeaves++;
	                return child;
	            }
	            startNode.numberOfLeaves++;
	            return search(child, suffix);
	        }
	    }
	    startNode.numberOfLeaves++;
	    return startNode;
	}
	
	public SuffixTreeNode find(SuffixTreeNode startNode, String suffix) {
		System.out.println(suffix);
	    if (suffix.isEmpty()) {
	        throw new IllegalArgumentException(
	                "Empty suffix. Probably no valid simple suffix tree exists for the input.");
	    }
	    Collection<SuffixTreeNode> children = startNode.children;
	    for (SuffixTreeNode child : children) {
		System.out.println(child.toString());
	        if (child.incomingEdge.label.charAt(0)==suffix.charAt(0)) {
	           suffix= suffix.substring(1);
	            if (suffix.isEmpty()) {
	                return child;
	            }
	            return find(child, suffix);
	        }
	    }
	    return null;
	}
	
	private void insert(SuffixTreeNode insertAt, String suffix,
	        int pathIndex) {
		char[] sArray=suffix.toCharArray();
	    for (char x : sArray) {
	        SuffixTreeNode child = new SuffixTreeNode(insertAt, x, insertAt.nodeDepth + 1, pathIndex, id);
	        insertAt.children.add(child);
	        insertAt = child;
	    }
	}
	
	public String toString() {
	    StringBuilder result = new StringBuilder();
	    String incomingLabel = this.isRoot() ? "" : this.incomingEdge.label;
	    for (int i = 1; i <= this.nodeDepth; i++)
	        result.append("\t");
	    if (this.isRoot()) {
	        c = 1;
	        this.id = 1;
	    } else {
	        this.id = c;
	        result.append(this.parent.id + " -> ");
	        result.append(this.id + "[label=\"" + incomingLabel + "\", numberOfLeaves="+this.numberOfLeaves+"];\n");
	    }
	    for (SuffixTreeNode child : children) {
	        c++;
	        child.id = c;
	        result.append(child.toString());
	    }
	    return result.toString();
	}

	
	public boolean isRoot() {
	    return this.parent == null;
	}
	public boolean isLeaf() {
	    return children.size() == 0;
	}
	
	
	public SuffixTreeNode traverseChar(int pos, char c){
		// reached end of edge
		if(this.incomingEdge==null ||this.incomingEdge.label.isEmpty() || pos+1>this.incomingEdge.label.length()-1){
			// find next edge
			for(SuffixTreeNode child:this.children){
				if(child.incomingEdge.label.charAt(0)==c){
					return child;
				}
			}
			// no edge found
			return null;
		}
		if(c==this.incomingEdge.label.charAt(pos+1)){
			return this;
		}
		return null;
	}
	
}

class SuffixTreeEdge {
	
	String label = null;
	@SuppressWarnings("unused")
	private int branchIndex = -1;

	public SuffixTreeEdge(String label, int branchIndex) {
	    this.label = label;
	    this.branchIndex = branchIndex;
	}
	
	public SuffixTreeEdge(char label, int branchIndex) {
	    this.label = label+"";
	    this.branchIndex = branchIndex;
	}

	
	
	
	
}
